@@ -1,43 +1,25 @@
 #!/usr/bin/python
 """Check the MANIFEST.in file in a Python source package for completeness.
 
-Here's the plan:
-    This script works by building a source distribution archive (by running
-    setup.py sdist), then checking the file list in the archive against the
-    file list in version control (Subversion, Git, Mercurial, Bazaar are
-    supported).
+This script works by building a source distribution archive (by running
+setup.py sdist), then checking the file list in the archive against the
+file list in version control (Subversion, Git, Mercurial, Bazaar are
+supported).
 
-    Since the first check can fail to catch missing MANIFEST.in entries when
-    you've got the right setuptools plugins installed, the script performs a
-    second test: unpacks the source distribution into a temporary directory,
-    then builds a second source distribution, and compares the file list again.
-
-    Alternatively it may be a better idea to export the source tree into a
-    temporary directory, build an sdist there, then compare it with the version
-    control list?
-
-Features currently implemented:
-
-    * getting file list from Subversion (executes svn in a subprocess)
-    * getting file list from Mercurial (executes hg in a subprocess)
-    * getting file list from Git (executes git in a subprocess)
-    * getting file list from Bazaar (executes bzr in a subprocess)
-    * comparing it with the list of files in a .tar.gz source distribution
-
-It is currently usable for checking if you can produce complete source
-distributions for uploading to PyPI, provided that your package lives in SVN.
-
-It's not usable for checking the completeness of a MANIFEST.in: the presence
-of the right setuptools plugin on your system might mean you're getting a
-complete sdist even without a complete MANIFEST.in.  (That's why the plan
-talks about a second sdist and/or VCS export.)
+Since the first check can fail to catch missing MANIFEST.in entries when
+you've got the right setuptools version control system support plugins
+installed, the script copies all the versioned files into a temporary
+directory before building the source distribution.  This also avoids
+issues with stale egg-info/SOURCES.txt files that may cause files not mentioned
+in MANIFEST.in to be included nevertheless.
 
 The current implementation probably doesn't work on Windows.
 """
 
 import argparse
-import re
+import fnmatch
 import os
+import re
 import shutil
 import subprocess
 import sys
@@ -47,7 +29,7 @@ import zipfile
 from contextlib import contextmanager
 
 
-__version__ = '0.8'
+__version__ = '0.9'
 __author__ = 'Marius Gedminas <marius@gedmin.as>'
 __licence__ = 'GPL v2 or later' # or ask me for MIT
 __url__ = 'https://github.com/mgedmin/check-manifest'
@@ -100,6 +82,11 @@ def error(message):
     print >> sys.stderr, message
 
 
+def warning(message):
+    _check_tbc()
+    print >> sys.stderr, message
+
+
 def format_list(list_of_strings):
     return "\n".join("  " + s for s in list_of_strings)
 
@@ -174,6 +161,24 @@ def mkdtemp(hint=''):
         shutil.rmtree(dirname)
 
 
+def copy_files(filelist, destdir):
+    """Copy a list of files to destdir, preserving directory structure.
+
+    File names should be relative to the current working directory.
+    """
+    for filename in filelist:
+        destfile = os.path.join(destdir, filename)
+        # filename should not be absolute, but let's double-check
+        assert destfile.startswith(destdir + os.path.sep)
+        destfiledir = os.path.dirname(destfile)
+        if not os.path.isdir(destfiledir):
+            os.makedirs(destfiledir)
+        if os.path.isdir(filename):
+            os.mkdir(destfile)
+        else:
+            shutil.copy2(filename, destfile)
+
+
 def get_one_file_in(dirname):
     """Return the pathname of the one file in a directory.
 
@@ -295,16 +300,30 @@ def add_directories(names):
 # Packaging logic
 #
 
-IGNORE = set([
-    'PKG-INFO',  # always generated
-    'setup.cfg', # always generated, sometimes also kept in source control
+# it's fine if any of these are missing in the VCS or in the sdist
+IGNORE = [
+    'PKG-INFO',     # always generated
+    '*.egg-info',   # always generated
+    '*.egg-info/*', # always generated
+    'setup.cfg',    # always generated, sometimes also kept in source control
     # it's not a problem if the sdist is lacking these files:
     '.hgtags', '.hgignore', '.gitignore', '.bzrignore',
-])
+    # it's convenient to ship compiled .mo files in sdists, but they shouldn't
+    # be checked in
+    '*.mo',
+]
+
+WARN_ABOUT_FILES_IN_VCS = [
+    # generated files should not be committed into the VCS
+    'PKG-INFO',
+    '*.egg-info',
+    '*.mo',
+]
 
 SUGGESTIONS = [(re.compile(pattern), suggestion) for pattern, suggestion in [
     # regexp -> suggestion
     ('^([^/]+[.](cfg|ini))$',       r'include \1'),
+    ('^([.]travis[.]yml)$',         r'include \1'),
     ('^([A-Z]+)$',                  r'include \1'),
     ('^[^/]+[.](txt|rst|py)$',      r'include *.\1'),
     ('^([a-zA-Z_][a-zA-Z_0-9]*)/'
@@ -316,23 +335,36 @@ SUGGESTIONS = [(re.compile(pattern), suggestion) for pattern, suggestion in [
 
 
 
+def file_matches(filename, patterns):
+    """Does this filename match any of the patterns?"""
+    return any(fnmatch.fnmatch(filename, pat) for pat in patterns)
+
+
 def strip_sdist_extras(filelist):
     """Strip generated files that are only present in source distributions."""
     return [name for name in filelist
-            if name not in IGNORE
-               and not name.endswith('.egg-info')
-               and '.egg-info/' not in name]
+            if not file_matches(name, IGNORE)]
+
+
+def find_bad_ideas(filelist):
+    """Find files matching WARN_ABOUT_FILES_IN_VCS patterns."""
+    return [name for name in filelist
+            if file_matches(name, WARN_ABOUT_FILES_IN_VCS)]
 
 
 def find_suggestions(filelist):
     """Suggest MANIFEST.in patterns for missing files."""
     suggestions = set()
+    unknowns = []
     for filename in filelist:
         for pattern, suggestion in SUGGESTIONS:
             m = pattern.match(filename)
             if m is not None:
                 suggestions.add(pattern.sub(suggestion, filename))
-    return sorted(suggestions)
+                break
+        else:
+            unknowns.append(filename)
+    return sorted(suggestions), unknowns
 
 
 def is_package(source_tree='.'):
@@ -349,30 +381,40 @@ def check_manifest(source_tree='.', create=False, update=False):
 
     Returns True if the manifest is fine.
     """
+    all_ok = True
     with cd(source_tree):
         if not is_package(source_tree):
             raise Failure('This is not a Python project (no setup.py).')
         info_begin("listing source files under version control")
-        source_files = sorted(strip_sdist_extras(get_vcs_files()))
+        all_source_files = sorted(get_vcs_files())
+        source_files = strip_sdist_extras(all_source_files)
         info_continue(": %d files and directories" % len(source_files))
-        info_begin("building an sdist")
-        with mkdtemp('-sdist') as tempdir:
-            run(['python', 'setup.py', 'sdist', '-d', tempdir])
-            sdist_filename = get_one_file_in(tempdir)
-            info_continue(": %s" % os.path.basename(sdist_filename))
-            sdist_files = sorted(strip_sdist_extras(strip_toplevel_name(
-                                    get_archive_file_list(sdist_filename))))
-            info_continue(": %d files and directories" % len(sdist_files))
-        if source_files != sdist_files:
+        info_begin("copying source files to a temporary directory")
+        with mkdtemp('-sources') as tempsourcedir:
+            copy_files(source_files, tempsourcedir)
+            info_begin("building an sdist")
+            with cd(tempsourcedir):
+                with mkdtemp('-sdist') as tempdir:
+                    run(['python', 'setup.py', 'sdist', '-d', tempdir])
+                    sdist_filename = get_one_file_in(tempdir)
+                    info_continue(": %s" % os.path.basename(sdist_filename))
+                    sdist_files = sorted(strip_sdist_extras(strip_toplevel_name(
+                                            get_archive_file_list(sdist_filename))))
+                    info_continue(": %d files and directories" % len(sdist_files))
+        if source_files == sdist_files:
+            info("files in version control match files in the sdist")
+        else:
             error("files in version control do not match the sdist!\n%s"
                   % format_difference(source_files, sdist_files,
                                       "VCS", "sdist"))
             missing_files = set(source_files) - set(sdist_files)
-            suggestions = find_suggestions(missing_files)
+            suggestions, unknowns = find_suggestions(missing_files)
+            user_asked_for_help = update or (create and not
+                                                os.path.exists('MANIFEST.in'))
             if suggestions:
                 info("suggested MANIFEST.in rules:\n%s"
                      % format_list(suggestions))
-                if update or (create and not os.path.exists('MANIFEST.in')):
+                if user_asked_for_help:
                     with open('MANIFEST.in', 'a') as f:
                         if f.tell() == 0:
                             info("creating MANIFEST.in")
@@ -380,10 +422,23 @@ def check_manifest(source_tree='.', create=False, update=False):
                             info("updating MANIFEST.in")
                             f.write('\n# added by check_manifest.py\n')
                         f.write('\n'.join(suggestions) + '\n')
-            return False
-        else:
-            info("files in version control match files in the sdist")
-        return True
+                    if unknowns:
+                        info("don't know how to come up with rules matching\n%s"
+                             % format_list(unknowns))
+            elif user_asked_for_help:
+                info("don't know how to come up with rules"
+                     " matching any of the files, sorry!")
+            all_ok = False
+        bad_ideas = find_bad_ideas(all_source_files)
+        if bad_ideas:
+            warning("you have %s in source control!\nthat's a bad idea:"
+                    " auto-generated files should not be versioned"
+                    % bad_ideas[0])
+            if len(bad_ideas) > 1:
+                warning("this also applies to the following:\n%s"
+                        % format_list(bad_ideas[1:]))
+            all_ok = False
+    return all_ok
 
 
 #
@@ -418,7 +473,7 @@ def zest_releaser_check(data):
     """Check the completeness of MANIFEST.in before the release.
 
     This is an entry point for zest.releaser.  See the documentation at
-    https://pypi.python.org/pypi/zest.releaser/3.3#entrypoints-documentation
+    http://zestreleaser.readthedocs.org/en/latest/entrypoints.html
     """
     from zest.releaser.utils import ask
     if not ask("Do you want to run check-manifest?"):
